package org.neuroph.contrib.tileclassification.gui;

import java.awt.AWTException;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.filechooser.FileFilter;

import org.neuroph.contrib.tileclassification.model.NeuralNetworkManager;
import org.neuroph.contrib.tileclassification.model.labels.AnalogLabelResponses;
import org.neuroph.contrib.tileclassification.model.labels.BinaryLabelResponses;
import org.neuroph.contrib.tileclassification.model.labels.LabelResponses;
import org.neuroph.contrib.tileclassification.model.plugin.TileClassificationPlugin;
import org.neuroph.contrib.tileclassification.screentools.FractionRgbData;
import org.neuroph.contrib.tileclassification.screentools.fileutils.ImagesLoader;
import org.neuroph.core.Layer;
import org.neuroph.core.NeuralNetwork;
import org.neuroph.core.Neuron;
import org.neuroph.core.learning.LearningRule;
import org.neuroph.nnet.learning.MomentumBackpropagation;
import org.neuroph.util.plugins.LabelsPlugin;

public class TileClassificationTrainerGui
{
	public static final String APPLICATION_TITLE = "Tile Classification Trainer";
	public static final int APPLICATION_WIDTH = 600;
	public static final int APPLICATION_HEIGHT = 600;
	
	private JFrame frame;
	private JLabel netErrorLabel;
	private TrainerSwingWorker trainerSwingWorker = null;
	private NeuralNetworkManager neuralNetworkManager;
	private JButton startLearningButton;
	private JButton stopLearningButton;
	private JFileChooser dirFileChooser;
	private JTextField dirOfJunkImages;
	private JTextField dirOfLabeledImages;
	private JTextField neuralNetworkName;
	private JButton saveNeuralNetworkButton;
	private JLabel topologyLabel;
	private JButton loadButton;
	private JTextArea outputLabelsTextArea;
	private JFileChooser imgfileChooser;
	private JButton chooseTestImageButton;
	
	// sampling resolution can be made dynamic later
	private Dimension samplingResolution = new Dimension(16, 16);
	private JTextField learningRateTextField;
	private JLabel testImageLabel;
	private JRadioButton binaryButton;
	private JRadioButton analogButton;
	private ButtonGroup labelResponsesButtonGroup;
	private ButtonGroup learningRuleButtonGroup;
	private JRadioButton momentumBackpropButton;
	private JRadioButton backpropButton;
	private ButtonGroup samplingResolutionButtonGroup;
	private JRadioButton sixteenButton;
	private JRadioButton thirtytwoButton;

	public static void main(String[] args) throws Exception
	{
		// Schedule a job for the event-dispatching thread:
		// creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			public void run()
			{
				try {
					new TileClassificationTrainerGui();
				} catch (Exception e) {
					e.printStackTrace();
					System.exit(0);
				}
			}
		});
	}

	public TileClassificationTrainerGui() throws AWTException, InterruptedException, IOException, ClassNotFoundException
	{
		neuralNetworkManager = new NeuralNetworkManager();
		
		try {
			UIManager.setLookAndFeel(
			        UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e1) {
			e1.printStackTrace();
		}
		
		frame = new JFrame(APPLICATION_TITLE);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		Container contentPane = frame.getContentPane();
		contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.Y_AXIS));
		
		populateContentPane(contentPane);

		// Display the window.
		frame.setPreferredSize(new Dimension(APPLICATION_WIDTH, APPLICATION_HEIGHT));
		frame.pack();
		frame.setVisible(true);
	}

	private void createTrainingThread()
	{
		stopTrainingThread();
		
		Map<String, FractionRgbData> rgbDataMap = new HashMap<String, FractionRgbData>();
		List<String> targetLabels = new ArrayList<String>();
		
		try {
			File labeledImagesDir = new File(dirOfLabeledImages.getText());
			rgbDataMap.putAll(ImagesLoader.getFractionRgbDataForDirectory(
					labeledImagesDir, samplingResolution));
			
			for(String imgName : rgbDataMap.keySet()) {
				StringTokenizer st = new StringTokenizer(imgName, "._");
				String target = st.nextToken();
				if(!targetLabels.contains(target)) {
					targetLabels.add(target);
				}
			}
			Collections.sort(targetLabels);
		} catch(IOException ioe) {
			System.err.println("Unable to load images from labeled images dir: '" + dirOfLabeledImages.getText() + "'");
			System.err.println(ioe.toString());
		}

		try {
			File junkImagesDir = new File(dirOfJunkImages.getText());
			rgbDataMap.putAll(ImagesLoader.getFractionRgbDataForDirectory(
					junkImagesDir, samplingResolution));
		} catch(IOException ioe) {
			System.err.println("Unable to load images from junk images dir: '" + dirOfJunkImages.getText() + "'");
			System.err.println(ioe.toString());
		}
		
		boolean useMomentumBackprop = false;
		String actionCommand = learningRuleButtonGroup.getSelection().getActionCommand();
		if("Momentum Backpropagation".equalsIgnoreCase(actionCommand)) {
			useMomentumBackprop = true;
		}
		
		if(needToCreateNewNetwork()) {
			LabelResponses labelResponses;
			actionCommand = labelResponsesButtonGroup.getSelection().getActionCommand();
			if("Analog".equalsIgnoreCase(actionCommand)) {
				labelResponses = new AnalogLabelResponses();
			} else {
				labelResponses = new BinaryLabelResponses();
			}
			
			neuralNetworkManager.createNewActiveClassificationNeuralNetwork(
					neuralNetworkName.getText(), samplingResolution, labelResponses, useMomentumBackprop, targetLabels, rgbDataMap);
		} else {
			neuralNetworkManager.createAndSetTrainingSet(useMomentumBackprop, targetLabels, rgbDataMap);
		}
		setTopologyAndLabels();
		
		trainerSwingWorker = new TrainerSwingWorker(this);
		trainerSwingWorker.execute();
	}
	
	private boolean needToCreateNewNetwork() {
		if(neuralNetworkManager.getActiveNeuralNetwork() == null) {
			return true;
		}
		
		NeuralNetwork activeNeuralNetwork = neuralNetworkManager.getActiveNeuralNetwork();
		
		TileClassificationPlugin tcPlugin = (TileClassificationPlugin) activeNeuralNetwork.getPlugin(TileClassificationPlugin.PLUGIN_NAME);
		Dimension currentSamplingResolution = tcPlugin.getSamplingResolution();
		
		String newName = neuralNetworkName.getText();
		LabelsPlugin labelsPlugin = (LabelsPlugin)activeNeuralNetwork.getPlugin(LabelsPlugin.LABELS_PLUGIN_NAME);
		String oldName = labelsPlugin.getLabel(activeNeuralNetwork);
		return !newName.equals(oldName) || !samplingResolution.equals(currentSamplingResolution);
	}

	private void stopTrainingThread()
	{
		if(trainerSwingWorker != null) {
			trainerSwingWorker.cancel(true);
		}
	}

	private void populateContentPane(Container contentPane)
	{
		JPanel namePanel = createXAxisBoxLayoutPanel(contentPane);
		JLabel neuralNetworkNameLabel = new JLabel("Neural Network Name");
		namePanel.add(neuralNetworkNameLabel);
		neuralNetworkName = new JTextField();
		namePanel.add(neuralNetworkName);
		
		loadButton = new JButton("Load From Disk");
		loadButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				try {
					loadNeuralNetwork();
				} catch (IOException e1) {
					e1.printStackTrace();
				} catch (ClassNotFoundException e1) {
					e1.printStackTrace();
				}
			}
		});
		namePanel.add(loadButton);
		
		labelResponsesButtonGroup = new ButtonGroup();
		
		binaryButton = new JRadioButton();
		binaryButton.setText("Binary");
		binaryButton.setActionCommand("Binary");
		labelResponsesButtonGroup.add(binaryButton);
		namePanel.add(binaryButton);
		
		analogButton = new JRadioButton();
		analogButton.setText("Analog");
		analogButton.setActionCommand("Analog");
		labelResponsesButtonGroup.add(analogButton);
		namePanel.add(analogButton);
		
		labelResponsesButtonGroup.setSelected(binaryButton.getModel(), true);
		
		
		JPanel samplingResolutionPanel = createXAxisBoxLayoutPanel(contentPane);
		
		JLabel label = new JLabel("Sampling Resolution: ");
		samplingResolutionPanel.add(label);
		
		samplingResolutionButtonGroup = new ButtonGroup();
		
		sixteenButton = new JRadioButton();
		sixteenButton.setText("16x16");
		sixteenButton.setActionCommand("16x16");
		sixteenButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				samplingResolution = new Dimension(16, 16);
			}
		});
		samplingResolutionButtonGroup.add(sixteenButton);
		samplingResolutionPanel.add(sixteenButton);
		
		thirtytwoButton = new JRadioButton();
		thirtytwoButton.setText("32x32");
		thirtytwoButton.setActionCommand("32x32");
		thirtytwoButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				samplingResolution = new Dimension(32, 32);
			}
		});
		samplingResolutionButtonGroup.add(thirtytwoButton);
		samplingResolutionPanel.add(thirtytwoButton);
		
		samplingResolutionButtonGroup.setSelected(sixteenButton.getModel(), true);
		
		
		
		dirFileChooser = new JFileChooser(new File("."));
		dirFileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		
		JPanel labeledImagesSelectionPanel = createXAxisBoxLayoutPanel(contentPane);
		dirOfLabeledImages = new JTextField();
		JButton directoryOfLabeledImagesButton = new JButton("Choose Directory of Labeled Images");
		directoryOfLabeledImagesButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				int returnVal = dirFileChooser.showDialog(null, "Select Directory");
				if(returnVal == JFileChooser.APPROVE_OPTION) {
					dirOfLabeledImages.setText(dirFileChooser.getSelectedFile().getAbsolutePath());
				}
			}
		});
		labeledImagesSelectionPanel.add(dirOfLabeledImages);
		labeledImagesSelectionPanel.add(directoryOfLabeledImagesButton);
		
		JPanel junkImagesSelectionPanel = createXAxisBoxLayoutPanel(contentPane);
		dirOfJunkImages = new JTextField();
		JButton directoryOfJunkButton = new JButton("Choose Directory of Junk Images");
		directoryOfJunkButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				int returnVal = dirFileChooser.showDialog(null, "Select Directory");
				if(returnVal == JFileChooser.APPROVE_OPTION) {
					dirOfJunkImages.setText(dirFileChooser.getSelectedFile().getAbsolutePath());
				}
			}
		});
		junkImagesSelectionPanel.add(dirOfJunkImages);
		junkImagesSelectionPanel.add(directoryOfJunkButton);

		JPanel learningControlsPanel = createXAxisBoxLayoutPanel(contentPane);
		JLabel learningRateLabel = new JLabel("Learning Rate:");
		learningControlsPanel.add(learningRateLabel);
		learningRateTextField = new JTextField();
		learningRateTextField.setText("0.10");
		learningControlsPanel.add(learningRateTextField);
		
		
		learningRuleButtonGroup = new ButtonGroup();
		
		momentumBackpropButton = new JRadioButton();
		momentumBackpropButton.setText("Momentum Backpropagation");
		momentumBackpropButton.setActionCommand("Momentum Backpropagation");
		learningRuleButtonGroup.add(momentumBackpropButton);
		learningControlsPanel.add(momentumBackpropButton);
		
		backpropButton = new JRadioButton();
		backpropButton.setText("Backpropagation");
		backpropButton.setActionCommand("Backpropagation");
		learningRuleButtonGroup.add(backpropButton);
		learningControlsPanel.add(backpropButton);
		
		learningRuleButtonGroup.setSelected(momentumBackpropButton.getModel(), true);
		
		
		startLearningButton = new JButton("Start Learning");
		startLearningButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				loadButton.setEnabled(false);
				learningRateTextField.setEditable(false);
				momentumBackpropButton.setEnabled(false);
				backpropButton.setEnabled(false);
				startLearningButton.setEnabled(false);
				stopLearningButton.setEnabled(true);
				saveNeuralNetworkButton.setEnabled(false);
				chooseTestImageButton.setEnabled(false);
				createTrainingThread();
			}
		});
		learningControlsPanel.add(startLearningButton);
		
		stopLearningButton = new JButton("Stop Learning");
		stopLearningButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				stopTrainingThread();
				loadButton.setEnabled(true);
				learningRateTextField.setEditable(true);
				momentumBackpropButton.setEnabled(true);
				backpropButton.setEnabled(true);
				startLearningButton.setEnabled(true);
				stopLearningButton.setEnabled(false);
				saveNeuralNetworkButton.setEnabled(true);
				chooseTestImageButton.setEnabled(true);
			}
		});
		stopLearningButton.setEnabled(false);
		learningControlsPanel.add(stopLearningButton);
		
		JPanel savePanel = createXAxisBoxLayoutPanel(contentPane);
		saveNeuralNetworkButton = new JButton("Save Neural Network");
		saveNeuralNetworkButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				try {
					// TODO: make the filepath dynamic, user specified
					neuralNetworkManager.persistNeuralNetwork(new File("."));
				} catch (IOException e1) {
					e1.printStackTrace();
				}
			}
		});
		saveNeuralNetworkButton.setEnabled(false);
		savePanel.add(saveNeuralNetworkButton);
		
		JPanel learningOutputPanel = createXAxisBoxLayoutPanel(contentPane);
		netErrorLabel = new JLabel("Net Error = ");
		learningOutputPanel.add(netErrorLabel);
		
		JPanel topologyPanel = createXAxisBoxLayoutPanel(contentPane);
		topologyLabel = new JLabel("Neural Network Topology: ");
		topologyPanel.add(topologyLabel);
		
		imgfileChooser = new JFileChooser(new File("."));
		imgfileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		imgfileChooser.setFileFilter(new FileFilter() {
			public boolean accept(File f) {
				if(f.isDirectory()) {
					return true;
				}
				String name = f.getName();
				if(name.endsWith(".png") || name.endsWith(".PNG") ||
						name.endsWith(".jpg") || name.endsWith(".JPG")) {
					return true;
				}
				return false;
			}
			public String getDescription() {
				// TODO Auto-generated method stub
				return null;
			}
		});
		
		JPanel testImageSelectionPanel = createXAxisBoxLayoutPanel(contentPane);
		chooseTestImageButton = new JButton("Choose Test Image");
		chooseTestImageButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				int returnVal = imgfileChooser.showDialog(null, "Select Image");
				if(returnVal == JFileChooser.APPROVE_OPTION) {
					try {
						BufferedImage img = ImagesLoader.loadImage(imgfileChooser.getSelectedFile(), samplingResolution);
						testImageLabel.setIcon(new ImageIcon(img));
						setOutputForLabels(img);
					} catch (IOException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}
				}
			}
		});
		testImageSelectionPanel.add(chooseTestImageButton);
		
		testImageLabel = new JLabel();
		testImageSelectionPanel.add(testImageLabel);
		
		
		JPanel outputLabelsPanel = createXAxisBoxLayoutPanel(contentPane);
		outputLabelsTextArea = new JTextArea(80, 10);
		outputLabelsTextArea.setEditable(false);
		outputLabelsTextArea.setAutoscrolls(true);
		JScrollPane pScroll = new JScrollPane(outputLabelsTextArea, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		outputLabelsPanel.add(pScroll);
	}
	
	protected void setOutputForLabels(BufferedImage img) {
		NeuralNetwork activeNeuralNetwork = getActiveNeuralNetwork();
		FractionRgbData rgbData = new FractionRgbData(img);
		activeNeuralNetwork.setInput(rgbData.getFlattenedRgbValues());
		activeNeuralNetwork.calculate();
		
		NumberFormat numberFormat = DecimalFormat.getNumberInstance();
		numberFormat.setMaximumFractionDigits(3);
		LabelsPlugin labelsPlugin = (LabelsPlugin) activeNeuralNetwork.getPlugin(LabelsPlugin.LABELS_PLUGIN_NAME);
		String neuralNetworkLabels = "";
		int i = 0;
		for(Neuron outputNeuron : activeNeuralNetwork.getOutputNeurons()) {
			String label = labelsPlugin.getLabel(outputNeuron);
			if(label == null) {
				label = "";
			}
			double output = outputNeuron.getOutput();
			neuralNetworkLabels += i + ": " + label + " = " + numberFormat.format(output) + "\n";
			i++;
		}
		outputLabelsTextArea.setText(neuralNetworkLabels);
	}

	protected void setNetError(double netError) {
		NumberFormat numberFormat = DecimalFormat.getNumberInstance();
		numberFormat.setMaximumFractionDigits(4);
		netErrorLabel.setText("Net Error = " + numberFormat.format(netError));
	}
	
	public void setTopologyAndLabels() {
		String layerNeuronCounts = "";
		NeuralNetwork activeNeuralNetwork = getActiveNeuralNetwork();
		for(Layer layer : activeNeuralNetwork.getLayers()) {
			layerNeuronCounts += layer.getNeuronsCount() + " ";
		}
		topologyLabel.setText("Neural Network Topology: " + layerNeuronCounts);
		
		LabelsPlugin labelsPlugin = (LabelsPlugin) activeNeuralNetwork.getPlugin(LabelsPlugin.LABELS_PLUGIN_NAME);
		String neuralNetworkLabels = "";
		int i = 0;
		for(Neuron outputNeuron : activeNeuralNetwork.getOutputNeurons()) {
			String label = labelsPlugin.getLabel(outputNeuron);
			if(label == null) {
				label = "";
			}
			neuralNetworkLabels += i + ": " + label + "\n";
			i++;
		}
		outputLabelsTextArea.setText(neuralNetworkLabels);
	}
	
	private JPanel createXAxisBoxLayoutPanel(Container contentPane)
	{
		JPanel panel = new JPanel();
		panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
		contentPane.add(panel);
		return panel;
	}
	
	public NeuralNetwork getActiveNeuralNetwork() {
		return neuralNetworkManager.getActiveNeuralNetwork();
	}
	
	public double getLearningRate() {
		try {
			return Double.parseDouble(learningRateTextField.getText());
		} catch(NumberFormatException nfe) {
			learningRateTextField.setText("0.25");
			return 0.25d;
		}
	}
	
	private void loadNeuralNetwork() 
		throws IOException, ClassNotFoundException 
	{
		String neuralNetworkNameStr = neuralNetworkName.getText();
		if ("".equals(neuralNetworkNameStr)) {
			throw new FileNotFoundException(
					"Neural network name must be specified!");
		}
		// TODO: make the filepath dynamic, user specified
		neuralNetworkManager.loadNeuralNetwork(new File("."),
				neuralNetworkNameStr);
		setTopologyAndLabels();
		setRadioButtons();
	}

	private void setRadioButtons() {
		TileClassificationPlugin tileClassificationPlugin = (TileClassificationPlugin)neuralNetworkManager.getActiveNeuralNetwork().getPlugin(TileClassificationPlugin.PLUGIN_NAME);
		if(tileClassificationPlugin == null) {
			System.err.println("TileClassificationPlugin was null for the loaded network!  Defaulting to BinaryLabelResponses");
			labelResponsesButtonGroup.setSelected(binaryButton.getModel(), true);
			samplingResolutionButtonGroup.setSelected(sixteenButton.getModel(), true);
		} else {
			if(tileClassificationPlugin.getLabelResponses() instanceof AnalogLabelResponses) {
				labelResponsesButtonGroup.setSelected(analogButton.getModel(), true);
			} else {
				labelResponsesButtonGroup.setSelected(binaryButton.getModel(), true);
			}
			
			if(tileClassificationPlugin.getSamplingResolution().equals(new Dimension(32, 32))) {
				samplingResolutionButtonGroup.setSelected(thirtytwoButton.getModel(), true);
			} else {
				samplingResolutionButtonGroup.setSelected(sixteenButton.getModel(), true);
			}
		}
		
		LearningRule learningRule = neuralNetworkManager.getActiveNeuralNetwork().getLearningRule();
		if(learningRule instanceof MomentumBackpropagation) {
			learningRuleButtonGroup.setSelected(momentumBackpropButton.getModel(), true);
		} else {
			learningRuleButtonGroup.setSelected(backpropButton.getModel(), true);
		}
	}
}
