package org.neuroph.contrib.tileclassification.screentools.fileutils;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.imageio.ImageIO;

public class ImagesIterator implements Iterator<BufferedImage>
{
	private Iterator<File> imageIterator;
	private String currentFilename = null;
	
	public ImagesIterator(File dir) throws IOException {
		if(!dir.isDirectory()) {
			throw new IOException(dir + " is not a directory!");
		}
		
		String[] jpgFilenames = dir.list(new FilenameFilter() {
			public boolean accept(File dir, String name)
			{
				if(name.length() > 4) {
    				String fileExtension = name.substring(name.length()-4, name.length());
    				return ".jpg".equalsIgnoreCase(fileExtension) ||
    						".png".equalsIgnoreCase(fileExtension);
				}
				return false;
			}
		});
		
		List<File> jpgFiles = new ArrayList<File>();
		for(String jpgFilename : jpgFilenames) {
			jpgFiles.add(new File(dir, jpgFilename));
		}
		
		imageIterator = jpgFiles.iterator();
	}

	public boolean hasNext()
	{
		return imageIterator.hasNext();
	}

	public BufferedImage next()
	{
		try {
			File next = imageIterator.next();
			currentFilename = next.getName();
			return ImageIO.read(next);
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	public void remove()
	{
		imageIterator.remove();
	}
	
	public String getFilenameOfCurrentImage() {
		return currentFilename;
	}
}
