package org.neuroph.contrib.tileclassification.screentools.fileutils;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

import javax.imageio.ImageIO;

import org.neuroph.contrib.tileclassification.screentools.FractionRgbData;
import org.neuroph.contrib.tileclassification.screentools.ScreenSampler;

public class ImagesLoader
{
	public static Map<String, FractionRgbData> getFractionRgbDataForDirectory(File imgDir, Dimension samplingResolution) throws IOException
	{
		if(!imgDir.isDirectory()) {
			throw new IOException("The given file must be a directory.  Argument is: " + imgDir);
		}
		
		Map<String, FractionRgbData> rgbDataMap = new HashMap<String, FractionRgbData>();
		
		ImagesIterator imagesIterator = new ImagesIterator(imgDir);
		while (imagesIterator.hasNext()) {
			BufferedImage img = imagesIterator.next();
			img = ScreenSampler.downSampleImage(samplingResolution, img);
			String filenameOfCurrentImage = imagesIterator.getFilenameOfCurrentImage();
//			System.out.println(filenameOfCurrentImage + " is: " + img.getWidth() + "x" + img.getHeight());
			StringTokenizer st = new StringTokenizer(filenameOfCurrentImage, ".");
			rgbDataMap.put(st.nextToken(), new FractionRgbData(img));
		}
		return rgbDataMap;
	}
	
	public static BufferedImage loadImage(File imgFile, Dimension samplingResolution) 
		throws IOException
	{
		BufferedImage img = ImageIO.read(imgFile);
		img = ScreenSampler.downSampleImage(samplingResolution, img);
		return img;
	}
}
