/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.neuroph.nnet.learning;

import org.neuroph.core.NeuralNetwork;
import org.neuroph.core.Neuron;
import org.neuroph.core.Connection;

/**
 * The OutstarLearning class implements hebbian-like learning rule for Outstar network.
 * @author Zoran Sevarac <sevarac@gmail.com>
 */
public class OutstarLearning extends UnsupervisedHebbianLearning {
	
	/**
	 * The class fingerprint that is set to indicate serialization
	 * compatibility with a previous version of the class.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Creates new instance of OutstarLearning algorithm  for the specified
	 * neural network.
	 * 
	 * @param neuralNetwork
     *                  neural network to train
	 */
	public OutstarLearning(NeuralNetwork neuralNetwork) {
		super(neuralNetwork);
		this.setLearningRate(0.1);
	}	
	
	/**
	 * This method implements weights update procedure for the single neuron
	 * 
	 * @param neuron
	 *            neuron to update weights for
	 */
	@Override
	protected void updateNeuronWeights(Neuron neuron) {
		double output = neuron.getOutput();
		for (Connection connection : neuron.getInputConnections()) {
			double input = connection.getInput();
			double weight = connection.getWeight().getValue();
			double deltaWeight = this.learningRate * input * (output-weight);
			connection.getWeight().inc(deltaWeight);
		}
	}	
	

}
